
// (c) COPYRIGHT URI/MIT 1994-1996
// Please read the full copyright statement in the file COPYRIGHT.
//
// Authors:
//      reza            Reza Nekovei (reza@intcomm.net)

// Implmentation for NCConnect: a subclass of Connect specialized for netcdf.
//
// jhrg 9/29/94

#include "config_nc.h"

static char rcsid[] not_used = {"$Id: NCConnect.cc,v 1.10 2003/05/02 16:30:29 jimg Exp $"};

#ifdef __GNUG__
#pragma implementation
#endif

#include <netcdf.h>

#if 0
#include "Connect.h"
#endif

#include "NCConnect.h"
#include "InternalErr.h"
#include "debug.h"

#ifdef WIN32
using std::cerr;
using std::endl;
#endif

// private mfuncs

void
NCConnect::init_list(int i)
{
    while(i--){
	_das_loc[i] = -1;
    }
}

// public mfuncs

NCConnect::NCConnect(const string &name) : AISConnect(name)
{
    if(is_local()){
	_ncid = -1;		// default value
    }
    else {
	_ncid = -1;
	_nvars = -1;
	_recdim = -1;
	init_list(MAX_NC_VARS); // clear the look up table
	// *** AcceptType is no longer used. 12/26/02 jhrg
#if 0
	if (get_accept_types() == "All")
	    set_accept_types("!Sequence");
#endif
    }
}	

NCConnect::~NCConnect()
{
    DBG(cerr << "Entering the NCConnect dtor" << endl);
    DBG(cerr << "Leaving the NCConnect dtor" << endl);
}

int
NCConnect::ncid()
{
    if (!is_local())
	throw InternalErr("ncid() is only valid for a local connection");

    return _ncid;
}

void
NCConnect::set_ncid(int id)
{
    if (!is_local())
	throw InternalErr("ncid() is only valid for a local connection");
    
    _ncid = id;
}

int
NCConnect::ndims()
{
    if (is_local())
	throw InternalErr("ndims() is only valid for a remote connection");

    return _ndims;
}

int
NCConnect::nvars()
{
    if (is_local())
	throw InternalErr("nvars() is only valid for a remote connection");

    return _nvars;
}


// Returns the dimension ID for the unlimited dimension
int
NCConnect::recdim()
{
    if (is_local())
	throw InternalErr("recdim() is only valid for a remote connection");

    return _recdim;
}


// Returns the dimension size for the given dimension ID
int
NCConnect::dim_size(int dimid)
{
    if (is_local())
	throw InternalErr("dim_size() is only valid for a remote connection");

    return _dim_size[dimid];
}


// Returns the dimension name for the given dimension ID
const string & 
NCConnect::dim_name(int dimid)
{
    if (is_local())
	throw InternalErr("dim_name() is only valid for a remote connection");

    return _dim_name[dimid];
}


// Returns the variable location in das table
int
NCConnect::das_loc(int varid)
{
    if (is_local())
	throw InternalErr("das_loc() is only valid for a remote connection");

    return _das_loc[varid];
}

// *** This function assumes that the variable with an unlimited dimension is
// at the top level (not inside a structure). If COARDS is adopted as a
// general convention, this code will need to be changed. 12/26/02 jhrg

// Set the ID for unlimited dimension from DODS_EXTRA attribute table
void
NCConnect::set_recdim()
{
    for (Pix q = das().first_var(); q; das().next_var(q))
	if (das().get_name(q) == "DODS_EXTRA") {  
	    string dim = das().get_table(q)->get_attr("Unlimited_Dimension");
	    for (int i = 0; i < _ndims; i++) {
		// DAS string has quotes
		string dim_name_tmp = "\"";
		dim_name_tmp += _dim_name[i] + "\"";
		if(dim_name_tmp == dim) {
		    _recdim = i;
		    break;
		}
	    }
	    break;
	}
}


// Creates a look up table indicating the variable locations in das table
// gived dds known order.
void
NCConnect::parse_das_loc()
{
    int var_tbl = 0, atr_tbl = 0, blank = 0;
    AttrTable *attr_table;

    // Add a blank table to point to for missing variable in DAS 
    attr_table = das().get_table("NC_BLANK");
    if (!attr_table)
      attr_table = das().add_table("NC_BLANK", new AttrTable);
    // Find its index 
    Pix q = das().first_var();
    for (blank=0; das().get_name(q) != "NC_BLANK" ; blank++)
      das().next_var(q);
    
    // Search and index all atributes for all the variables
    for (Pix p = Cdds().first_var(); p; atr_tbl=0, var_tbl++, Cdds().next_var(p))
	for (q = das().first_var(); q; atr_tbl++, das().next_var(q))
	    if (Cdds().var(p)->name() == das().get_name(q)) {
		_das_loc[var_tbl+1]=atr_tbl;  
		break;
	    }
    _nvars = var_tbl;
    // Reserve the first location for NC_GLOBAL (which has a varid = -1)
    // or any other GLOBAL atrribute (ie. MAT_GLOBAL, HDF_GLOBAL, DSP_GLOBAL,
    // etc.)
    atr_tbl = 0;
    for (q = das().first_var(); q; atr_tbl++, das().next_var(q))
      if((das().get_name(q)).find("_GLOBAL") != string ::npos){
	_das_loc[0]=atr_tbl;  
	break;
      }

    // Set index to point at the blank attribute table for variables without it
    for(q = das().first_var(), var_tbl=0; q ; var_tbl++, das().next_var(q))
      if(_das_loc[var_tbl] == -1)
	_das_loc[var_tbl]= blank;    
}

// Adds dimensions to the look up table for Grid class 
void
NCConnect::parse_grid_dims()
{
    int j;
    string vartype, mapdimname;
    Grid *gr;
    Array *ar;

    // Grid type: 
                          
    //search all the variables
    for (Pix p = Cdds().first_var(); p;  Cdds().next_var(p)){ 
	vartype = Cdds().var(p)->type_name();               

	// find Grids
	if(vartype ==  "Grid"){                   
	    gr =  (Grid *) Cdds().var(p);

	    // Get all its Array maps
	    for (Pix q = gr->first_map_var(); q; gr->next_map_var(q)){
		ar = (Array *)gr->map_var(q);                                 

		// Get dimension name
		mapdimname = ar->dimension_name(ar->first_dim()).c_str();
		if (mapdimname == "") // if dim. name is missing use var. name 
		    mapdimname = ar->name();

		// Find a match to the dimension name and size ?
		for(j=0 ; j <_ndims; j++)
		    if((_dim_name[j] ==  mapdimname)&&        
		       (_dim_size[j] == ar->dimension_size(ar->first_dim())))
			break;

		// No match found, add it to the dimension tables
		if(j == _ndims){
		    _ndims++;
		    _dim_name[j] =  mapdimname;
		    _dim_size[j] = ar->dimension_size(ar->first_dim());
		}
	    }
	}
    }
}


// Adds dimensions to the look up table for Array class 
void
NCConnect::parse_array_dims()
{
    int j, dim_cnt;
    string vartype, dimname;
    Array *ar;
    char dim_rep[2];

    // Array type:

    //search all the variables
    for (Pix p = Cdds().first_var(); p;  Cdds().next_var(p)){
	vartype = Cdds().var(p)->type_name();

	// find Array
	if(vartype == "Array"){
	    ar =  (Array *) Cdds().var(p);

	    // search all dimensions
	    dim_cnt = 0;
	    for (Pix d = ar->first_dim(); d;  ar->next_dim(d)){

		// get dimension name
		dimname = ar->dimension_name(d);

		// Find a match to a dimension size (and name if it exist) ?
		for(j=0 ; j <_ndims; j++)
		    if(dimname == "") {

			if(_dim_size[j] == ar->dimension_size(d))
			    break;

		    }
		    else {

			if((_dim_name[j] == dimname) &&
			   (_dim_size[j] == ar->dimension_size(d)))
			    break;
		    }

		// No match found, add it to the dimension tables
		if(j == _ndims){
		    if (dimname == ""){
			sprintf(dim_rep,"%d",dim_cnt);
			dimname = ar->name() + (string)dim_rep;
		    }
		    _dim_name[j+dim_cnt] = dimname;
		    _dim_size[j+dim_cnt] = ar->dimension_size(d);
		    dim_cnt++;
		}
	    }
	    _ndims  += dim_cnt; // All the dimensions in a variable must be diff.
	}
    }
}


// Creates a look up table for dimensions names and sizes
// Note:  Currently only handles Grid and Array types.
void
NCConnect::parse_dims()
{
    _ndims=0;

    parse_grid_dims();
    parse_array_dims();

    set_recdim(); 

#ifdef TEST
    cerr << _ndims << _recdim << endl;
    for(int i=0; i<_ndims;i++)
	cerr << _dim_name[i] << " " <<_dim_size[i] <<endl;
#endif
}

// When adding the AIS functions to Connect, I (finally) removed the DAS
// field and request_das() method that 'held state.' These two methods are
// here to make up for that lost interface. Ideally they should be
// replaced/removed and the code reworked. 04/15/03 jhrg
void
NCConnect::request_das()
{
    AISConnect::request_das(_das);
}

DAS &
NCConnect::das()
{
    return _das;
}

// Gets the constarined DDS from the server side
// without having to get the data
void
NCConnect::request_Cdds(const string &expr, const string &)
{
    request_dds(_Cdds, expr);
}

DDS &
NCConnect::Cdds()
{
    return _Cdds;
}

// $Log: NCConnect.cc,v $
// Revision 1.10  2003/05/02 16:30:29  jimg
// Fixes for the builds
//
// Revision 1.9  2003/03/06 01:41:06  jimg
// Removed calls to get/set_accept_types().
//
// Revision 1.8  2003/01/28 07:08:24  jimg
// Merged with release-3-2-8.
//
// Revision 1.4.2.9  2002/12/27 00:34:51  jimg
// Added the set_ncid() method. I added this because I'm tracking down a bug and
// what to be sure that the code isn't modifying the internal value of _ncid
// somewhere (which it could do if it has access to a reference to the field).
//
// Revision 1.4.2.8  2002/12/27 00:05:36  jimg
// I changed a number of the accessors here to throw InternalErr instead of
// writing to stderr.
//
// Revision 1.4.2.7  2002/12/23 23:06:40  jimg
// Build Fixes.
//
// Revision 1.4.2.6  2002/09/08 23:58:35  rmorris
// Added a "return" statement in a function call.  VC++ requires.
//
// Revision 1.4.2.5  2002/06/27 21:00:26  jimg
// Fixed call to Connect::request_dds(...). The local method request_Cdds(...)
// uses Connect's new version of request_dds(...). This file was 'fixed' once to
// use my first try at a new interface for Connect. I then fixed Connect some
// more and broke this code again... I think the current interface supported by
// Connect is much better.
//
// Revision 1.4.2.4  2002/06/21 00:31:40  jimg
// I changed many files throughout the source so that the 'make World' build
// works with the new versions of Connect and libdap++ that use libcurl.
// Most of these changes are either to Makefiles, configure scripts or to
// the headers included by various C++ files. In a few places the oddities
// of libwww forced us to hack up code and I've undone those and some of the
// clients had code that supported libwww's generous tracing capabilities
// (that's one part of libwww I'll miss); I had to remove support for that.
// Once this code compiles and more work is done on Connect, I'll return to
// each of these changes and polish them.
//
// Revision 1.7  2002/05/03 00:01:52  jimg
// Merged with release-3-2-7.
//
// Revision 1.4.2.3  2001/12/26 03:44:04  rmorris
// Make it use cerr, endl from the std namespace under WIN32.
//
// Revision 1.6  2001/09/28 17:18:41  jimg
// Merged with 3.2.5.
// CVS  Committing in .
//
// Revision 1.4.2.2  2001/09/27 06:00:37  jimg
// Addd debug.h
//
// Revision 1.5  2001/08/30 23:08:24  jimg
// Merged with 3.2.4
//
// Revision 1.4.2.1  2001/08/30 06:23:03  reza
// Fixed bugs reported by bugzilla.
//
// Revision 1.4  2000/10/30 17:28:12  jimg
// Fixes for exception-based error reporting.
//
// Revision 1.3  2000/10/06 01:22:02  jimg
// Moved the CVS Log entries to the ends of files.
// Modified the read() methods to match the new definition in the dap library.
// Added exception handlers in various places to catch exceptions thrown
// by the dap library.
//
// Revision 1.2  1999/11/05 05:15:05  jimg
// Result of merge woth 3-1-0
//
// Revision 1.1.2.1  1999/10/29 05:05:21  jimg
// Reza's fixes plus the configure & Makefile update
//
// Revision 1.1  1999/07/28 00:22:42  jimg
// Added
//
// Revision 1.17  1999/05/07 23:45:31  jimg
// String --> string fixes
//
// Revision 1.16  1999/04/09 17:56:23  jimg
// Added call to set_accept_types telling servers that we don't like
// Sequences.
//
// Revision 1.15  1998/08/06 16:33:22  jimg
// Fixed misuse of the read(...) member function. Return true if more data
// is to be read, false is if not and error if an error is detected
//
// Revision 1.14  1998/02/05 20:14:46  jimg
// DODS now compiles with gcc 2.8.x
//
// Revision 1.13  1997/03/26 05:44:22  reza
// GLOBAL attribute can now be named anything with "_GLOBAL" in it (i.e.
// HDF_GLOBAL, MAT_GLOBAL, DSP_GLOBAL, etc.). This should improve
// interoperability with non-netCDF servers. Also, the DAS no longer requires
// attributes for all the variables. An internal empty attribute table
// (NC_BLANK) manages missing ones.
//
// Revision 1.12  1997/02/12 19:14:11  reza
// Upgraded for release 2.11.
//
// Revision 1.11  1996/11/13 22:23:07  jimg
// Added debuging to NCConnect dtor
//
// Revision 1.10  1996/09/17 17:06:22  jimg
// Merge the release-2-0 tagged files (which were off on a branch) back into
// the trunk revision.
//
// Revision 1.9.2.3  1996/09/17 00:26:21  jimg
// Merged changes from a side branch which contained various changes from
// Reza and Charles.
// Removed ncdump and netexec since ncdump is now in its own directory and
// netexec is no longer used.
//
// Revision 1.9.2.2  1996/07/10 21:44:00  jimg
// Changes for version 2.06. These fixed lingering problems from the migration
// from version 1.x to version 2.x.
// Removed some (but not all) warning generated with gcc's -Wall option.
//
// Revision 1.9.2.1  1996/06/25 22:04:23  jimg
// Version 2.0 from Reza.
//
// Revision 1.10  1995/11/18  11:33:22  reza
// Updated member function names for DAP-1.1.1.
//
// Revision 1.9  1995/11/10  06:31:22  reza
// Removed API name from connect.
//
// Revision 1.8  1995/07/09  21:33:42  jimg
// Added copyright notice.
//
// Revision 1.7  1994/12/22  04:51:42  reza
// Updated to use DODS new named dimension capability.
//
// Revision 1.6  1994/11/23  20:59:52  reza
// Added heuristic dimension matching for Grid and Array classes.
// Added Dimension handling routines (ndims(), dim_size(), dim_name(),
// parse_dim(),and etc.)
//
// Revision 1.5  1994/11/18  21:33:33  reza
// Added _nvars and nvars() for number of variables in the file.
// Added _das_loc[] and parse_das_loc() to make a look-up table for variables
// in dds vs. das.
//
// Revision 1.4  1994/11/03  05:45:07  reza
// Moved error, request_das, and request _dds to the surrogate library for more
// flexibility and simpler error handling.
//
// Revision 1.3  1994/10/28  19:09:03  jimg
// Added _error, error() and fixed ncid() so that it retruns a reference.
//
// Revision 1.2  1994/10/05  20:23:30  jimg
// Fixed errors in *.h files comments - CVS bites again.
// Changed request_{das,dds} so that they use the field `_api_name'
// instead of requiring callers to pass the api name.
//
// Revision 1.1  1994/10/05  18:02:14  jimg
// First version of the connection management classes.
// This commit also includes early versions of the test code.
