/*
 *	Copyright 1996, University Corporation for Atmospheric Research
 *      See netcdf/COPYRIGHT file for copying and redistribution conditions.
 * $Id: lv2i.c,v 1.3 2002/05/03 00:01:52 jimg Exp $ */

#include "lnc.h"

#ifndef NO_NETCDF_2

#if SIZEOF_LONG == SIZEOF_SIZE_T
/*
 * We don't have to copy the arguments to switch from 'long'
 * to 'size_t' or 'ptrdiff_t'. Use dummy macros.
 */

# define NDIMS_DECL
# define A_DECL(name, type, ndims, rhs) \
	const type *const name = ((const type *)(rhs))

# define A_FREE(name)

# define A_INIT(lhs, type, ndims, rhs)
	
#else 
/*
 * We do have to copy the arguments to switch from 'long'
 * to 'size_t' or 'ptrdiff_t'. In my tests on an SGI,
 * any additional cost was lost in measurement variation.
 */

# include "lonstack.h"

static size_t
nvdims(int ncid, int varid)
{
	NC *ncp;
	if(NC_check_id(ncid, &ncp) != NC_NOERR)
		return 0;
	{
		const NC_var *const varp = NC_lookupvar(ncp, varid);
		if(varp == NULL)
			return 0;
		return varp->ndims;
	}
}

#define NDIMS_DECL	const size_t ndims = nvdims(ncid, varid);

# define A_DECL(name, type, ndims, rhs) \
	ALLOC_ONSTACK(name, type, ndims)

# define A_FREE(name) \
	FREE_ONSTACK(name)

# define A_INIT(lhs, type, ndims, rhs) \
	{ \
		const long *lp = rhs; \
		type *tp = lhs; \
		type *const end = lhs + ndims; \
		while(tp < end) \
		{ \
			*tp++ = (type) *lp++; \
		} \
	}


#endif


/* Begin globals */

/*
 * Error code
 */
int ncerr = NC_NOERR ;


/*
 * The subroutines in error.c emit no messages unless NC_VERBOSE bit is on.
 * They call exit() when NC_FATAL bit is on.
 */
int ncopts = (NC_FATAL | NC_VERBOSE) ;


/*
 * Backward compatibility for the version 2 fortran jackets
 */
const char *cdf_routine_name;


/* End globals */

/* Begin error handling */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>


/*
 */
void
lnc_advise(const char *routine_name, int err, const char *fmt,...)
{
    va_list args;

    if(NC_ISSYSERR(err))
	ncerr = NC_SYSERR;
    else
	ncerr = err;

    if( ncopts & NC_VERBOSE )
	{
	    (void) fprintf(stderr,"%s: ", routine_name);
	    va_start(args ,fmt);
	    (void) vfprintf(stderr,fmt,args);
	    va_end(args);
	    if(err != NC_NOERR)
		{
		    (void) fprintf(stderr,": %s",
				   lnc_strerror(err));
		}
	    (void) fputc('\n',stderr);
	    (void) fflush(stderr);	/* to ensure log files are current */
	}

    if( (ncopts & NC_FATAL) && err != NC_NOERR )
	{
	    exit(ncopts);
	}
}

/* 
   Backward compatibility for the version 2 fortran jackets

   Note: I've scanned the source code and I cannot find where this function
   is called in the fortran jacket code. However, I do see where this
   function is called by our C++ source files Dattr.cc, etc. Also, I see that
   the fortran jackets calls nc_advise (look in fort-v2compat.c). 8/23/99
   jhrg

   Note 2: I found that as of 3.2.1 (but probably earlier) the error messages
   produced by this code were garbled. The params sent in using the varying
   args got trashed. I *think* that this was caused by calling va_start(args)
   twice on the same parameters. So, even though code duplication is a BAD
   THING(tm), I copied the stuff from lnc_advise() to the NCadvise() and
   nc_advise(). This fixes the error messages produced by NCadvise. Here's
   an example:

   old:  ncopen: Error in getting DAS from ': Invalid argument
   new: ncopen: Error in getting DAS from
   `http://localhost/dods-test/nph-dods/data/nc/fnoc5.nc?u': Invalid argument

   I claim it's an improvement... 4/17/2002 jhrg
*/

void
NCadvise(int err, char *fmt,...)
{
    va_list args;

    if(NC_ISSYSERR(err))
	ncerr = NC_SYSERR;
    else
	ncerr = err;

    if( ncopts & NC_VERBOSE ) {
	(void) fprintf(stderr,"%s: ", cdf_routine_name);
	va_start(args ,fmt);
	(void) vfprintf(stderr,fmt,args);
	va_end(args);
	if(err != NC_NOERR)
	    {
		(void) fprintf(stderr,": %s",
			       lnc_strerror(err));
	    }
	(void) fputc('\n',stderr);
	(void) fflush(stderr);	/* to ensure log files are current */
    }

    if( (ncopts & NC_FATAL) && err != NC_NOERR) {
	exit(ncopts);
    }
}

/*
  I'm adding this function because Kevin O'Brien (@ PMEL) has reported that
  fortran programs won't link because a definition for it is missing. It
  seems that the real netCDF 3.x library defines this, so maybe we should be
  adding this in the upper (nc3-dods) directory and not here in lnetcdf.
  8/23/99 jhrg 
*/
void
nc_advise(const char *routine_name, int err, const char *fmt,...)
{
    va_list args;
    if(NC_ISSYSERR(err))
	ncerr = NC_SYSERR;
    else
	ncerr = err;

    if( ncopts & NC_VERBOSE ) {
	(void) fprintf(stderr,"%s: ", cdf_routine_name);
	va_start(args ,fmt);
	(void) vfprintf(stderr,fmt,args);
	va_end(args);
	if(err != NC_NOERR)
	    {
		(void) fprintf(stderr,": %s",
			       lnc_strerror(err));
	    }
	(void) fputc('\n',stderr);
	(void) fflush(stderr);	/* to ensure log files are current */
    }

    if( (ncopts & NC_FATAL) && err != NC_NOERR) {
	exit(ncopts);
    }
#if 0
    va_start(args ,fmt);
    lnc_advise(routine_name, err, fmt, args);
    va_end(args);
#endif
}


/* End error handling */


int
lnccreate(const char* path, int cmode)
{
	int ncid;
	const int status = lnc_create(path, cmode, &ncid);
	if(status != NC_NOERR)
	{
		lnc_advise("lnccreate", status, "filename \"%s\"", path);
		return -1;
	}
	return ncid;
}


int
lncopen(const char *path, int mode)
{
	int ncid;
	const int status = lnc_open(path, mode, &ncid);
	if(status != NC_NOERR)
	{
		lnc_advise("lncopen", status, "filename \"%s\"", path);
		return -1;
	}
	return ncid;
}


int
lncredef(int ncid)
{
	const int status =  lnc_redef(ncid);
	if(status != NC_NOERR)
	{
		lnc_advise("lncredef", status, "ncid %d", ncid);
		return -1;
	}
	return 0;
}


int
lncendef(int ncid)
{
	const int status = lnc_enddef(ncid);
	if(status != NC_NOERR)
	{
		lnc_advise("lncendef", status, "ncid %d", ncid);
		return -1;
	}
	return 0;
}


int
lncclose(int ncid)
{
	const int status = lnc_close(ncid);
	if(status != NC_NOERR)
	{
		lnc_advise("lncclose", status, "ncid %d", ncid);
		return -1;
		
	}
	return 0;
}


int
lncinquire(
    int		ncid,
    int*	ndims,
    int*	nvars,
    int*	natts, 
    int*	recdim
)
{
	int nd, nv, na;
	const int status = lnc_inq(ncid, &nd, &nv, &na, recdim);

	if(status != NC_NOERR)
	{
		lnc_advise("lncinquire", status, "ncid %d", ncid);
		return -1;
	}
	/* else */

	if(ndims != NULL)
		*ndims = (int) nd;

	if(nvars != NULL)
		*nvars = (int) nv;

	if(natts != NULL)
		*natts = (int) na;

	return ncid;
}


int
lncsync(int ncid)
{
	const int status = lnc_sync(ncid);
	if(status != NC_NOERR)
	{
		lnc_advise("lncsync", status, "ncid %d", ncid);
		return -1;
		
	}
	return 0;
}


int
lncabort(int ncid)
{
	const int status = lnc_abort(ncid);
	if(status != NC_NOERR)
	{
		lnc_advise("lncabort", status, "ncid %d", ncid);
		return -1;
	}
	return 0;
}


int
lncdimdef(
    int		ncid,
    const char*	name,
    long	length
)
{
	int dimid;
	const int status =  lnc_def_dim(ncid, name, (size_t)length, &dimid);
	if(status != NC_NOERR)
	{
		lnc_advise("lncdimdef", status, "ncid %d", ncid);
		return -1;
	}
	return dimid;
}


int
lncdimid(int ncid, const char*	name)
{
	int dimid;
	const int status =  lnc_inq_dimid(ncid, name, &dimid);
	if(status != NC_NOERR)
	{
		lnc_advise("lncdimid", status, "ncid %d", ncid);
		return -1;
	}
	return dimid;
}


int
lncdiminq(
    int		ncid,
    int		dimid,
    char*	name,
    long*	length
)
{
	size_t ll;
	const int status = lnc_inq_dim(ncid, dimid, name, &ll);

	if(status != NC_NOERR)
	{
		lnc_advise("lncdiminq", status, "ncid %d", ncid);
		return -1;
	}
	/* else */
	
	if(length != NULL)
		*length = (int) ll;

	return dimid;
}


int
lncdimrename(
    int		ncid,
    int		dimid,
    const char*	name
)
{
	const int status = lnc_rename_dim(ncid, dimid, name);
	if(status != NC_NOERR)
	{
		lnc_advise("lncdimrename", status, "ncid %d", ncid);
		return -1;
	}
	return dimid;
}


int
lncvardef(
    int		ncid,
    const char*	name,
    nc_type	datatype, 
    int		ndims,
    const int*	dim
)
{
	int varid = -1;
	const int status = lnc_def_var(ncid, name, datatype, ndims, dim, &varid);
	if(status != NC_NOERR)
	{
		lnc_advise("lncvardef", status, "ncid %d", ncid);
		return -1;
	}
	return varid;
}


int
lncvarid(
    int		ncid,
    const char*	name
)
{
	int varid = -1;
	const int status = lnc_inq_varid(ncid, name, &varid);
	if(status != NC_NOERR)
	{
		lnc_advise("lncvarid", status, "ncid %d", ncid);
		return -1;
	}
	return varid;
}


int
lncvarinq(
    int		ncid,
    int		varid,
    char*	name,
    nc_type*	datatype,
    int*	ndims,
    int*	dim,
    int*	natts
)
{
	int nd, na;
	const int status = lnc_inq_var(ncid, varid, name, datatype,
		 &nd, dim, &na);

	if(status != NC_NOERR)
	{
		lnc_advise("lncvarinq", status, "ncid %d", ncid);
		return -1;
	}
	/* else */
	
	if(ndims != NULL)
		*ndims = (int) nd;

	if(natts != NULL)
		*natts = (int) na;

	return varid;
}


int
lncvarput1(
    int		ncid,
    int		varid,
    const long*	index,
    const void*	value
)
{
	NDIMS_DECL
	A_DECL(coordp, size_t, ndims, index);
	A_INIT(coordp, size_t, ndims, index);
	{
	const int status = lnc_put_var1(ncid, varid, coordp, value);
	A_FREE(coordp);
	if(status != NC_NOERR)
	{
		lnc_advise("lncvarput1", status, "ncid %d", ncid);
		return -1;
	}
	}
	return 0;
}


int
lncvarget1(
    int		ncid,
    int		varid,
    const long*	index,
    void*	value
)
{
	NDIMS_DECL
	A_DECL(coordp, size_t, ndims, index);
	A_INIT(coordp, size_t, ndims, index);
	{
	const int status = lnc_get_var1(ncid, varid, coordp, value);
	A_FREE(coordp);
	if(status != NC_NOERR)
	{
		lnc_advise("lncdimid", status, "ncid %d", ncid);
		return -1;
	}
	}
	return 0;
}


int
lncvarput(
    int		ncid,
    int		varid,
    const long*	start,
    const long*	count, 
    const void*	value
)
{
	NDIMS_DECL
	A_DECL(stp, size_t, ndims, start);
	A_DECL(cntp, size_t, ndims, count);
	A_INIT(stp, size_t, ndims, start);
	A_INIT(cntp, size_t, ndims, count);
	{
	const int status = lnc_put_vara(ncid, varid, stp, cntp, value);
	A_FREE(cntp);
	A_FREE(stp);
	if(status != NC_NOERR)
	{
		lnc_advise("lncvarput", status, "ncid %d", ncid);
		return -1;
	}
	}
	return 0;
}


int
lncvarget(
    int		ncid,
    int		varid,
    const long*	start,
    const long*	count, 
    void*	value
)
{
	NDIMS_DECL
	A_DECL(stp, size_t, ndims, start);
	A_DECL(cntp, size_t, ndims, count);
	A_INIT(stp, size_t, ndims, start);
	A_INIT(cntp, size_t, ndims, count);
	{
	const int status = lnc_get_vara(ncid, varid, stp, cntp, value);
	A_FREE(cntp);
	A_FREE(stp);
	if(status != NC_NOERR)
	{
		lnc_advise("lncvarget", status, "ncid %d", ncid);
		return -1;
	}
	}
	return 0;
}


int
lncvarputs(
    int		ncid,
    int		varid,
    const long*	start,
    const long*	count,
    const long*	stride,
    const void*	value
)
{
	if(stride == NULL)
		return lncvarput(ncid, varid, start, count, value);
	/* else */
	{
	NDIMS_DECL
	A_DECL(stp, size_t, ndims, start);
	A_DECL(cntp, size_t, ndims, count);
	A_DECL(strdp, ptrdiff_t, ndims, stride);
	A_INIT(stp, size_t, ndims, start);
	A_INIT(cntp, size_t, ndims, count);
	A_INIT(strdp, ptrdiff_t, ndims, stride);
	{
	const int status = lnc_put_vars(ncid, varid, stp, cntp, strdp, value);
	A_FREE(strdp);
	A_FREE(cntp);
	A_FREE(stp);
	if(status != NC_NOERR)
	{
		lnc_advise("lncvarputs", status, "ncid %d", ncid);
		return -1;
	}
	}
	return 0;
	}
}


int
lncvargets(
    int		ncid,
    int		varid,
    const long*	start,
    const long*	count,
    const long*	stride,
    void*	value
)
{
	if(stride == NULL)
		return lncvarget(ncid, varid, start, count, value);
	/* else */
	{
	NDIMS_DECL
	A_DECL(stp, size_t, ndims, start);
	A_DECL(cntp, size_t, ndims, count);
	A_DECL(strdp, ptrdiff_t, ndims, stride);
	A_INIT(stp, size_t, ndims, start);
	A_INIT(cntp, size_t, ndims, count);
	A_INIT(strdp, ptrdiff_t, ndims, stride);
	{
	const int status = lnc_get_vars(ncid, varid, stp, cntp, strdp, value);
	A_FREE(strdp);
	A_FREE(cntp);
	A_FREE(stp);
	if(status != NC_NOERR)
	{
		lnc_advise("lncvargets", status, "ncid %d", ncid);
		return -1;
	}
	}
	return 0;
	}
}


int
lncvarputg(
    int		ncid,
    int		varid,
    const long*	start,
    const long*	count,
    const long*	stride,
    const long*	map,
    const void* value
)
{
	if(map == NULL)
		return lncvarputs(ncid, varid, start, count, stride, value);
	/* else */
	{
	NDIMS_DECL
	A_DECL(stp, size_t, ndims, start);
	A_DECL(cntp, size_t, ndims, count);
	A_DECL(strdp, ptrdiff_t, ndims, stride);
	A_DECL(imp, ptrdiff_t, ndims, map);
	A_INIT(stp, size_t, ndims, start);
	A_INIT(cntp, size_t, ndims, count);
	A_INIT(strdp, ptrdiff_t, ndims, stride);
	A_INIT(imp, ptrdiff_t, ndims, map);
	{
	const int status = lnc_put_varm(ncid, varid,
			 stp, cntp, strdp, imp, value);
	A_FREE(imp);
	A_FREE(strdp);
	A_FREE(cntp);
	A_FREE(stp);
	if(status != NC_NOERR)
	{
		lnc_advise("lncvarputg", status, "ncid %d", ncid);
		return -1;
	}
	}
	return 0;
	}
}


int
lncvargetg(
    int		ncid,
    int		varid,
    const long*	start,
    const long*	count,
    const long*	stride,
    const long*	map,
    void*	value
)
{
	if(map == NULL)
		return lncvargets(ncid, varid, start, count, stride, value);
	/* else */
	{
	NDIMS_DECL
	A_DECL(stp, size_t, ndims, start);
	A_DECL(cntp, size_t, ndims, count);
	A_DECL(strdp, ptrdiff_t, ndims, stride);
	A_DECL(imp, ptrdiff_t, ndims, map);
	A_INIT(stp, size_t, ndims, start);
	A_INIT(cntp, size_t, ndims, count);
	A_INIT(strdp, ptrdiff_t, ndims, stride);
	A_INIT(imp, ptrdiff_t, ndims, map);
	{
	const int status = lnc_get_varm(ncid, varid,
			stp, cntp, strdp, imp, value);
	A_FREE(imp);
	A_FREE(strdp);
	A_FREE(cntp);
	A_FREE(stp);
	if(status != NC_NOERR)
	{
		lnc_advise("lncvargetg", status, "ncid %d", ncid);
		return -1;
	}
	}
	return 0;
	}
}


int
lncvarrename(
    int		ncid,
    int		varid,
    const char*	name
)
{
	const int status = lnc_rename_var(ncid, varid, name);
	if(status != NC_NOERR)
	{
		lnc_advise("lncvarrename", status, "ncid %d", ncid);
		return -1;
	}
	return varid;
}


int
lncattput(
    int		ncid,
    int		varid,
    const char*	name, 
    nc_type	datatype,
    int		len,
    const void*	value
)
{
	const int status = lnc_put_att(ncid, varid, name, datatype, len, value);
	if(status != NC_NOERR)
	{
		lnc_advise("lncattput", status, "ncid %d", ncid);
		return -1;
	}
	return 0;
}


int
lncattinq(
    int		ncid,
    int		varid,
    const char*	name, 
    nc_type*	datatype,
    int*	len
)
{
	size_t ll;
	const int status = lnc_inq_att(ncid, varid, name, datatype, &ll);
	if(status != NC_NOERR)
	{
		lnc_advise("lncattinq", status, "ncid %d", ncid);
		return -1;
	}
	
	if(len != NULL)
		*len = (int) ll;

	return 1;

}


int
lncattget(
    int		ncid,
    int		varid,
    const char*	name, 
    void*	value
)
{
	const int status = lnc_get_att(ncid, varid, name, value);
	if(status != NC_NOERR)
	{
		lnc_advise("lncattget", status, "ncid %d", ncid);
		return -1;
	}
	return 1;
}


int
lncattcopy(
    int		ncid_in,
    int		varid_in,
    const char*	name, 
    int		ncid_out,
    int		varid_out
)
{
	const int status = lnc_copy_att(ncid_in, varid_in, name, ncid_out, varid_out);
	if(status != NC_NOERR)
	{
		lnc_advise("lncattcopy", status, "%s", name);
		return -1;
	}
	return 0;
}


int
lncattname(
    int		ncid,
    int		varid,
    int		attnum,
    char*	name
)
{
	const int status = lnc_inq_attname(ncid, varid, attnum, name);
	if(status != NC_NOERR)
	{
		lnc_advise("lncattname", status, "ncid %d", ncid);
		return -1;
	}
	return attnum;
}


int
lncattrename(
    int		ncid,
    int		varid,
    const char*	name, 
    const char*	newname
)
{
	const int status = lnc_rename_att(ncid, varid, name, newname);
	if(status != NC_NOERR)
	{
		lnc_advise("lncattrename", status, "ncid %d", ncid);
		return -1;
	}
	return 1;
}


int
lncattdel(
    int		ncid,
    int		varid,
    const char*	name
)
{
	 const int status = lnc_del_att(ncid, varid, name);
	if(status != NC_NOERR)
	{
		lnc_advise("lncattdel", status, "ncid %d", ncid);
		return -1;
	}
	return 1;
}

#endif /* NO_NETCDF_2 */

/*
 *  This is how much space is required by the user, as in
 *
 *   vals = malloc(nel * lnctypelen(var.type));
 *   lncvarget(cdfid, varid, cor, edg, vals);
 */
int
lnctypelen(nc_type type) 
{
	switch(type){
	case NC_BYTE :
	case NC_CHAR :
		return((int)sizeof(char));
	case NC_SHORT :
		return(int)(sizeof(short));
	case NC_INT :
		return((int)sizeof(int));
	case NC_FLOAT :
		return((int)sizeof(float));
	case NC_DOUBLE : 
		return((int)sizeof(double));
	}

#ifndef NO_NETCDF_2
	/* else */
	lnc_advise("lnctypelen", NC_EBADTYPE, "Unknown type %d",
		(int)type);
#endif /* NO_NETCDF_2 */
	return -1;
}

#ifndef NO_NETCDF_2

int
lncsetfill(
    int		ncid,
    int		fillmode
)
{
	int oldmode = -1;
	const int status = lnc_set_fill(ncid, fillmode, &oldmode);
	if(status != NC_NOERR)
	{
		lnc_advise("lncsetfill", status, "ncid %d", ncid);
		return -1;
	}
	return oldmode;
}


int
lncrecinq(
    int		ncid,
    int*	nrecvars,
    int*	recvarids,
    long*	recsizes
)
{
	size_t nrv = 0;
	size_t rs[NC_MAX_VARS]; /* TODO */
	const int status = lnc_inq_rec(ncid, &nrv, recvarids, rs);
	if(status != NC_NOERR)
	{
		lnc_advise("lncrecinq", status, "ncid %d", ncid);
		return -1;
	}

	if(nrecvars != NULL)
		*nrecvars = (int) nrv;

	if(recsizes != NULL)
	{
		size_t ii;
		for(ii = 0; ii < nrv; ii++)
		{
			recsizes[ii] = (long) rs[ii];
		}
	}

	return (int) nrv;
}


int
lncrecget(
    int		ncid,
    long	recnum,
    void**	datap
)
{
	const int status = lnc_get_rec(ncid, (size_t)recnum, datap);
	if(status != NC_NOERR)
	{
		lnc_advise("lncrecget", status, "ncid %d", ncid);
		return -1;
	}
	return 0;
}


int
lncrecput(
    int		ncid,
    long	recnum,
    void* const* datap
)
{
	const int status = lnc_put_rec(ncid, (size_t)recnum, datap);
	if(status != NC_NOERR)
	{
		lnc_advise("lncrecput", status, "ncid %d", ncid);
		return -1;
	}
	return 0;
}

#endif /* NO_NETCDF_2 */
